import decimal
import json
import uuid

from base64 import b64encode, b64decode

from django.db import DEFAULT_DB_ALIAS, router, models
from django.db.models import signals
from django.db.models.sql.where import WhereNode, AND
from django.utils.functional import Promise


class CompositeKey(dict):

    @staticmethod
    def encode(value: dict) -> str:
        return b64encode(json.dumps(value).encode('utf-8')).decode('utf-8')

    @staticmethod
    def decode(value: str) -> dict:
        return json.loads(b64decode(value.encode('utf-8')).decode('utf-8'))

    def to_json_string(self):
        return json.dumps(self)

    def __str__(self):
        return self.encode(self)

    def __hash__(self):
        return hash(tuple(self[key] for key in sorted(self.keys())))


class CompositeKeyField(models.AutoField):

    SUPPORTED_SUBFIELD_TYPES = (
        models.BooleanField,
        models.CharField,
        models.DecimalField,
        models.IntegerField,
        models.TextField,
        models.UUIDField,
        # We rely on foreign key being used only
        # to refer columns of supported types
        models.ForeignKey,
    )

    def __init__(self, fields: list[models.Field], **kwargs):
        for field in fields:
            self._validate_field(field)
        self.fields = fields
        self.columns = [field.db_column or field.name for field in fields]
        super().__init__(primary_key=True, **kwargs)

    def _validate_field(self, field: models.Field):
        if not isinstance(field, self.SUPPORTED_SUBFIELD_TYPES):
            raise TypeError(
                f'{field.name}: {field.get_internal_type()} cannot be used '
                f'as a subfield for {self.__class__.__name__}'
            )

    def contribute_to_class(self, cls, name, private_only=False):
        self.set_attributes_from_name(name)
        self.model = cls
        self.concrete = False
        self.editable = False
        self.column = self.columns[0]  # for default order_by
        cls._meta.add_field(self, private=True)
        cls._meta.setup_pk(self)

        if not getattr(cls, self.attname, None):
            setattr(cls, self.attname, self)

        def delete(inst, using=None, keep_parents=False):
            using = using or router.db_for_write(self.model, instance=inst)

            signals.pre_delete.send(
                sender=cls, instance=inst, using=using
            )

            query = cls._default_manager.filter(**self.__get__(inst))
            query._raw_delete(using)

            for column in self.columns:
                setattr(inst, column, None)

            signals.post_delete.send(
                sender=cls, instance=inst, using=using
            )

        cls.delete = delete

    def get_prep_value(self, value):
        return self.to_python(value)

    def to_python(self, value):
        if value is None or isinstance(value, CompositeKey):
            return value
        return CompositeKey(CompositeKey.decode(value))

    def to_json(self, value):
        if isinstance(value, (decimal.Decimal, uuid.UUID, Promise)):
            return str(value)
        return value

    def bulk_related_objects(self, objs, using=DEFAULT_DB_ALIAS):
        return []

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        return CompositeKey({
            column: self.to_json(self.model._meta.get_field(column).value_from_object(instance))
            for column in self.columns
        })

    def __set__(self, instance, value):
        pass


@CompositeKeyField.register_lookup
class Exact(models.Lookup):

    lookup_name = 'exact'

    def as_sql(self, compiler, connection):
        fields = [
            self.lhs.field.model._meta.get_field(column)
            for column in self.lhs.field.columns
        ]
        lookup_classes = [field.get_lookup('exact') for field in fields]
        lookups = [
            lookup_class(field.get_col(self.lhs.alias), self.rhs[column])
            for lookup_class, field, column in zip(lookup_classes, fields, self.lhs.field.columns)
        ]
        value_constraint = WhereNode()
        for lookup in lookups:
            value_constraint.add(lookup, AND)

        return value_constraint.as_sql(compiler, connection)



import logging

from enum import Enum
from functools import partial

from django.db import models
from sqlalchemy import Table, Column, types
from sqlalchemy.dialects.postgresql import UUID

from sqlalchemy_django_admin import fields


logger = logging.getLogger(__name__)


SA_TYPE_TO_MODEL = {
    types.String: models.CharField,
    types.Integer: models.IntegerField,
    types.DateTime: models.DateTimeField,
    types.Date: models.DateField,
    types.DECIMAL: models.DecimalField,
    types.Boolean: models.BooleanField,
    types.BigInteger: models.BigIntegerField,
    types.Text: models.TextField,
    types.Enum: models.CharField,
    UUID: models.UUIDField,
    # Use `json` field instead of `jsonb` for PostgreSQL
    types.JSON: fields.RawJSONField,
}


def _generate_model_name(table: Table):
    # Currently the name of the table is used as the corresponding model name
    return table.name


def _generate_model_verbose_name(table: Table):
    return ' '.join(part.capitalize() for part in table.name.split('_'))


def _convert_default(default):
    """
    Converts SQLAlchemy's default to Django supported default .
    Supports only small scope of SQLAlchemy's functionality:
    - https://docs.sqlalchemy.org/en/14/core/defaults.html#scalar-defaults
    - https://docs.sqlalchemy.org/en/14/core/defaults.html#python-executed-functions
    - enum
    """
    default = default and default.arg
    if default:
        if isinstance(default, Enum):
            default = default.value
        if callable(default):
            default = partial(default, None)
    return default


def _column_as_field(column: Column, pk_column: Column = None) -> models.Field:
    """
    Converts SQLAlchemy Column to Django Model Field
    """
    field_class = SA_TYPE_TO_MODEL[type(column.type)]
    kwargs = dict(
        name=column.name,
        null=column.nullable,
        default=_convert_default(column.default),
        # Primary key is calculated implicitly or can be set explicitly via `pk_column`.
        # Make sure `pk_column` refers to a column with unique values
        primary_key=column.primary_key or column is pk_column,
        # By default only nullable fields marked as not required
        blank=column.nullable,
    )

    if column is column.table.primary_key._autoincrement_column:
        field_class = models.AutoField

    if isinstance(column.type, types.Enum):
        kwargs['choices'] = [(i, i) for i in column.type.enums]

    if isinstance(column.type, types.String):
        if column.type.length:
            kwargs['max_length'] = column.type.length
        else:
            field_class = SA_TYPE_TO_MODEL[types.Text]

    if column.foreign_keys:
        # Trying to find the original table column which is referred from here.
        # Way to deal with composite fk (composite pk reference)
        foreign_key = None
        related_column = column
        while related_column.foreign_keys:
            # FIXME: what do we do if there are more than 1 fk?
            foreign_key, *_ = related_column.foreign_keys
            related_column = foreign_key.column

        # Making django foreign key only if the column refers
        # to the only primary key of related table
        is_unique_pk = (
            related_column.primary_key
            and len(related_column.table.primary_key.columns) == 1
        )
        if is_unique_pk:
            field_class = fields.ForeignKey
            related_model_name = _generate_model_name(related_column.table)
            kwargs |= dict(
                # Act as Django when name ends with `_id`
                name=column.name[:-3] if column.name.endswith('_id') else f'{column.name}_obj',
                to=f'sqlalchemy_django_admin.{related_model_name}',
                on_delete=models.PROTECT,
                to_field=related_column.name,
                db_column=column.name,
            )

    return field_class(**kwargs)


def table_as_model(
    table: Table,
    name: str = None,
    name_plural: str = None,
    str_template: str = None,
    pk_column: Column = None,
) -> type[models.Model]:
    """
    Converts SQLAlchemy Table to Django Model
    """
    model_name = _generate_model_name(table)
    model_verbose_name = name or _generate_model_verbose_name(table)
    model_fields = [_column_as_field(column, pk_column) for column in table.columns]

    # Deal with composite primary key
    primary_fields = [field for field in model_fields if field.primary_key]
    if len(primary_fields) > 1:
        for field in primary_fields:
            field.primary_key = False
        model_fields.append(fields.CompositeKeyField(primary_fields, name='composite_id'))

    class Meta:
        db_table = table.name
        managed = False
        app_label = 'sqlalchemy_django_admin'
        verbose_name = model_verbose_name
        verbose_name_plural = name_plural or f'{verbose_name}s'

    def __str__(obj):
        pk = obj.pk.to_json_string() if isinstance(obj.pk, fields.CompositeKey) else obj.pk
        template = str_template or '{model_verbose_name} object({pk})'
        return template.format(self=obj, pk=pk, model_verbose_name=model_verbose_name)

    attrs = {field.name: field for field in model_fields}
    attrs['Meta'] = Meta
    attrs['__str__'] = __str__
    attrs['__module__'] = 'sqlalchemy_django_admin'

    return type(model_name, (models.Model,), attrs)



from django.contrib import admin
from django.db.models import Field, BooleanField

from sqlalchemy_django_admin.fields import CompositeKeyField


# TODO: move to settings
DEFAULT_COLUMNS_COUNT = 4


class ModelAdmin(admin.ModelAdmin):
    """
    Adds some changes to default behavior of admin.ModelAdmin:
    1. Foreign keys are automatically added to `raw_id_fields`
       to prevent loading of all related objects on the form page.
    2. If `list_display` is not defined the first DEFAULT_COLUMNS_COUNT fields
       are shown on the list page instead of `__str__()` of the model.
    3. If `search_fields` is not defined it will search by the exact value
       of the primary key. Doesn't work for composite primary keys.
    4. If `list_filter` is not defined it will be filled with fields that have choices
       and fields of boolean type.
    5. All primary key fields (except auto fields) by default are shown on creation page
       but hidden on edition page.
    """
    @property
    def _model_fields(self) -> list[Field]:
        return [f for f in self.opts.get_fields() if f.concrete]

    @property
    def _primary_field_names(self) -> set[str]:
        if isinstance(self.opts.pk, CompositeKeyField):
            return {field.name for field in self.opts.pk.fields}
        return {self.opts.pk.name}

    @property
    def raw_id_fields(self):
        return [f.name for f in self.opts.get_fields() if f.is_relation]

    def get_list_display(self, request):
        if self.list_display == ('__str__',):
            return tuple(
                f.db_column if f.is_relation else f.name
                for f in self._model_fields[:DEFAULT_COLUMNS_COUNT]
            )
        return super().get_list_display(request)

    def get_search_fields(self, request):
        if not self.search_fields:
            return ['=pk']
        return super().get_search_fields(request)

    def get_list_filter(self, request):
        if not self.list_filter:
            return [f.name for f in self._model_fields if f.choices or isinstance(f, BooleanField)]
        return super().get_list_filter(request)

    def get_fields(self, request, obj=None):
        """
        In addition to default behavior, removes from edition page all primary key fields.
        PK has to be immutable, because if you change it,
        Django will just create a new object on .save()

        Django has similar model field parameter for this – `editable=False`,
        but it also removes a field from creation page.
        Thus it can't be automatically used for fields which must be set manually.
        """
        fields = super().get_fields(request, obj)
        if not obj:
            return fields
        return [field for field in fields if field not in self._primary_field_names]



from django.db import models

from .composite import CompositeKey, CompositeKeyField


class ForeignKey(models.ForeignKey):

    def get_attname(self):
        return self.db_column or super().get_attname()


class RawJSONField(models.JSONField):
    """
    Json field that works with `json` type columns in admin forms
    instead of `jsonb` for PostgreSQL
    """
    # Only null is invalid when blank=False
    empty_values = [None]

    def db_type(self, connection):
        return 'json'

    def from_db_value(self, value, expression, connection):
        try:
            return super().from_db_value(value, expression, connection)
        except TypeError:
            return value



from pathlib import Path
from setuptools import setup, find_packages

this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()

setup(
    name='sqlalchemy-django-admin',
    description='Django Admin for SQLAlchemy',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='kartashov',
    version='0.0.5',
    packages=find_packages(),
    install_requires=[
        'django>=4.0',
        'sqlalchemy>=1.4',
    ],
)
